# /r2/sketch_talk_classes.rb

$sketch_talk_fps = 24 # default frames per second

# SketchTalk Classes ------------------------------------------

# typing savers

class GT < Geom::Transformation
end

class P3d < Geom::Point3d
end

class V3d < Geom::Vector3d
end

# other classes

class Box
=begin
A box is a rectangular parallelepiped: six faces, all rectangular, all meeting at right angles. A box is defined by near and far corners of a rectangle plus a distance to PushPull the rectangle, creating the box shape. If the rectangle is orthogonal, a positive PushPull is done toward the positive end of the perpendicular axis. All six faces of the box will face the outside of the box.
=end
    attr_reader :rectangle, :pushpull_distance

    def initialize( near, far, pushpull_distance )     
        @rectangle = Rectangle.new( near, far )
        @pushpull_distance = pushpull_distance
    end
    
    def draw
        @rectangle.draw()
        @rectangle.face().pushpull( pushpull_distance )
    end
    
end # of class Box

class Circle
    attr_reader :nsides, :center, :radius, :normal, :edges, :face, :group
    
    def initialize( nsides, center, radius, plane_or_normal )
        @nsides = nsides
        @center = center
        @radius = radius
        @plane_or_normal = plane_or_normal
        @normal = make_normal( plane_or_normal )
    end # of initialize()
    
    def draw( ents )
        @group = ents.add_group()
        ents = @group.entities()
        @edges = ents.add_circle( @center, @normal, @radius, @nsides )
        @face = ents.add_face( @edges )
        orient( @face )
        ents.add_face( @edges )
        Sketchup.active_model().selection().add( @group )
        return @group
    end # of draw()
    
    def inspect()
        ret = self.to_s() + @nsides.to_s() + ' sides, center=' + (center.inspect) + ', radius='+@radius.to_s()
    end # of inspect()
        
    
end # of class Circle

class Column
=begin
A Column is a Circle PushPulled to a given height.
=end
    attr_reader :circle, :height

    def initialize( nsides, center, radius, plane_or_normal, height )
        @nsides = nsides
        @center = center
        @radius = radius
        @plane_or_normal = plane_or_normal
        @normal = make_normal( @plane_or_normal )
        @height = height
    end # of initialize()
    
    def draw( ents )
        @circle = Circle.new( @nsides, @center, @radius, @normal ) 
            # returns a Circle
        @circle.draw( ents )
        @circle.face().pushpull( @height )
    end # of draw()

end # of class Column

# Con_xxx classes are support classes for the Conductor

class Con_Amci < MovableCI
    # Amci = Actionable, movable component instance
    attr_reader :act
    
    def initialize( registrant, r, g, b )
        super( registrant.actionable )
        @registrant = registrant
        @r = r
        @g = g
        @b = b
    end
    
    def act( sec, frame )
        move_tw( @r, @g, @b )
    end
    
    def inspect()
        ret = '{Con_Amci: ' + @registrant.name + ' ' + @vector.inspect() + '}'
        return ret
    end
end

class Con_Calls
    attr_reader :registrants
    attr_accessor :stop_time

    def initialize( stop_time, registrants )
        @stop_time = stop_time
        @registrants = registrants
    end
    
    def call( second, frame )
        for reg in @registrants do reg.actionable.act( second, frame ) end
    end # of call()
    
    def inspect() 
        ret = '{Con_Calls: stop=' + @stop_time.to_s() +
            ' Con_Registrants: '
        for reg in @registrants do ret += reg.name + ',' end
        ret = ret.chop + '}'
        return ret
    end # of inspect()
    
end # of class Con_Calls

class Con_Registrant

    attr_reader :name, :vector, :start, :stop
    attr_accessor :actionable

    def initialize( actionable, name, vector, start, stop )
        @actionable = actionable
        @name = name
        @vector = vector
        @start = start
        @stop = stop
    end # of initialize()
    
    def wants_call?( sec )
        return ( sec >= @start ) && ( sec < @stop )
    end
    
    def inspect()
        return '{Con_Registrant: name=' + @name + 
            ', start=' + @start.to_s() + 
            ', stop=' + @stop.to_s() + '}'
    end # of inspect()
    
end # of class Con_Registrant

class Conductor
=begin
The Conductor is the view's animation. His job is to add registrants that wish to move (translate, rotate, scale) to his lists. As seconds pass, the Conductor will call the act() method of the registrants that want notification that second. (What the registrants act() methods do is unknown to the Conductor.)

The Conductor organizes the registrants into an array of Con_Calls objects. The Con_Calls object holds a stop time and a list of registrants to call. It also has a call() method the conductor will call until its stop time is reached. The call() method calls the individual registrants. On reaching the Con_Calls stop time, the Conductor moves on to the next Con_Calls object.

The goal here is to do the hard work before beginning the animation. Then call the appropriate registrants with minimal overhead. There are 40+ millis per frame. We want to leave as many millis as possible for the registrants to do their thing(s).
=end

    def initialize()
        @@call_list = []
        @@registrants = []
        @@num_reg = 0
        @@frame_num = -1 
        @@delay = 1.0 / ( $sketch_talk_fps + 6 )
        @@second = -1
        @@frames_per_sec = $sketch_talk_fps.to_i()
        @@call_ptr = 0
    end # of initialize()
    
    def nextFrame( view )
        
        @@frame_num += 1
        if ( @@frame_num % @@frames_per_sec ) == 0
            @@second += 1
            if @@call_list[ @@call_ptr ].stop_time == @@second
                @@call_ptr += 1
            end
        end
        return false if @@call_ptr == @@call_list.length()
        call = @@call_list[@@call_ptr]
        call.call( @@second, @@frame_num )
        
        view.show_frame( @@delay )
        return true #
    end # of nextFrame()
    
    def find_last()
        last = 0
        for reg in @@registrants do last = [ last, reg.stop ].max end
        return last
    end
    
    def create_amcis() # convert registrants into actionable, movable CIs
        for reg in @@registrants
            duration = reg.stop - reg.start
            frames = duration * $sketch_talk_fps
            r = reg.vector[0].to_f() / frames
            g = reg.vector[1].to_f() / frames
            b = reg.vector[2].to_f() / frames
            reg.actionable = Con_Amci.new( reg, r, g, b )
        end # registrant loop
    end # of create_amcis()
    
    def vec_div( vec, denom ) # divide array by scalar
        ret = []
        for v in vec do ret.push( (v.to_f( )) / denom ) end
        return ret
    end # of vec_div
    
    def organize() # converts list of Con_Registrants to list of Con_Calls
        call_ptr = 0
        @@call_list.push( Con_Calls.new(1, []) )
        last_sec = find_last() - 1
        old_list = []
        
        for sec in 0..last_sec
            list = []
            for reg in @@registrants
                list.push( reg ) if reg.wants_call?( sec ) 
            end
            
            next_sec = sec+1
            if list == old_list
                @@call_list[ call_ptr ].stop_time = next_sec
            else
                @@call_list.push( Con_Calls.new(next_sec, list) )
                call_ptr += 1
                old_list = list
            end
        end # seconds loop
        
    end # of organize()

    def register( args )
        # args: [instance ref, vector, start, stop]
        name = 'R' + @@num_reg.to_s()
        @@num_reg += 1
        reg = Con_Registrant.new( args[0], name, args[1], args[2], args[3] )
        @@registrants.push( reg )
        return true
    end # of register()
    
    def start()
        create_amcis()
        organize()
        model = Sketchup.active_model()
        view = model.active_view()
        view.animation = self
    end # of start()
    
    def stop()
        # any cleanup? doubt it, but just in case
    end # of stop()
    
end # of class Conductor

class Doughnut
=begin
By hand: drag a Rectangle from near to far. oFFset inside the rectangle. Delete the face in the center. PushPull the remaining face.
=end
    attr_reader :rectangle, :pushpull_distance, :thickness

    def initialize( near, far, pushpull_distance, thickness )
=begin
Parameters near and far are [r,g,b] arrays. The thickness is the distance from the outside to the hole.
=end    
        
        @rectangle = Rectangle.new( near, far )
        @pushpull_distance = pushpull_distance
        @thickness = thickness
        
    end # of initialize()
    
    def draw()
        @rectangle.draw()
        inner_rec = inset( @rectangle.near, @rectangle.far, @thickness )
        inner_rec.draw( @rectangle.group().entities() )
        inner_rec.group().explode()
        inner_rec.face().erase!()
        @rectangle.face().pushpull( @pushpull_distance )
    end # of draw()

    def inset( near, far, thickness )
=begin
        Computes new near/far pair, oFFset inside by thickness.
        This is messy as there is no guarantee that the coordinates
        of "near" are smaller than the coordinates of "far".
=end
        near2 = []
        far2 = []
        
        # First, assign smaller+thickness to near2, larger-thickness to far2.
        for i in 0..2
            arr = ins( near[i], far[i] )
            near2[i] = arr[0]
            far2[i] = arr[1]
        end

        # The coordinates of the perpendicular value V are now V+thickness and
        # V-thickness. ([0,0,0],[100,100,0] is now [8,8,8], [92,92,-8]).
        # Adjust them back to V and V.
        unless @plane == 'no'
            case @rectangle.plane()
                when 'rg' then i = 2
                when 'rb' then i = 1
                when 'gb' then i = 0
            end
            
            near2[i] = far2[i] = near[i] 
        end # adjust orthogonal planes

        rec = Rectangle.new( near2, far2 )
        return rec
        
    end # of inset()
    
    def ins( v1, v2 )
        if v1 < v2
            return [ v1+@thickness, v2-@thickness ] 
        else
            return [ v2+@thickness, v1-@thickness ]
        end 
    end # ins()

end # of class Doughnut

class Rectangle
=begin
A Rectangle is created from a near corner and a far corner. 
The corners are [r,g,b] arrays. 

An orthogonal rectangle is created, if possible. 
The rectangle's @plane is one of 'rg', 'gb', 'br' or 'no' (Not Orthogonal). 

An orthogonal Rectangle's normal points toward the positive end of the
perpendicular axis. 

A non-orthogonal Rectangle's outside face will face the positive end 
of the blue axis, except that Rectangles parallel to the blue axis face the positive end of the green axis.
=end
    attr_reader :near, :far, :plane, :face, :group
    
    def initialize( near, far )
        
        r = 0; g = 1; b = 2;
        @near = near; @far = far
        
        if near[r] == far[r]
            @plane = 'gb'
        elsif near[g] == far[g]
            @plane = 'rb'
        elsif near[b] == far[b]
            @plane = 'rg'
        else
            @plane = 'no'
        end
        
    end # of initialize()

    def draw( *args ) # args == none or optional group
        corners = get_corners()
        if args.length == 1
            corners.insert( 0, args[0] )
        end

        @group = add_face_group( corners )
        @face = find_face( @group.entities )
        return @group
        
    end # of draw()
    
    def move( vec )
        @near[0] += vec[0]; @near[1] += vec[1]; @near[2] += vec[2]
        @far[0]  += vec[0]; @far[1]  += vec[1]; @far[2]  += vec[2]
        group.entities.erase_entities( group.entities.to_a() )
        draw()
    end # of move()
    
    def inspect()
        return 'Rectangle: @near=' + @near.inspect() + ', @far=' + @far.inspect() 
    end
    
    def get_corners()

        r = 0; g = 1; b = 2;
        ret = []
        ret[0] = @near
        ret[2] = @far
        if ( @plane == 'rg' ) || ( @plane == 'no' )
            ret[1] = [ @near[r], @far[g], @far[b] ]
            ret[3] = [ @far[r], @near[g], @near[b] ]
        elsif @plane == 'gb'
            ret[1] = [ @near[r], @near[g], @far[b] ]
            ret[3] = [ @near[r], @far[g], @near[b] ]
        else # @plane == 'rb'
            ret[1] = [ @near[r], @near[g] ,@far[b] ]
            ret[3] = [ @far[r], @near[g], @near[b] ]
        end

        return ret

    end # get_corners()
    
    def center()
        ret = [ (@near[0]+@far[0])/2.0, 
            (@near[1]+@far[1])/2.0,
            (@near[2]+@far[2])/2.0 ]
        return ret
    end # of center()
    

end # of class Rectangle

class Timer
=begin
Written and placed in public domain by Martin Rinehart, January, 2010.

Thinking back to my Java days, I thought that UI.start/stop_timer() methods should really be replaced by a Timer class. You can do three things with a timer:
 
timer = Timer.new( pause_length, runnable )
timer.start()
timer.stop()

A Timer is initialized with a pause_length (seconds, including fractions of a second, often 1.0/frames_per_second) and a Runnable object. An object is Runnable if it has a run() method that returns a boolean (true==keep running, false==stop running). 

After the Timer's start() method is called, the Runnable's run() is called every pause_length seconds until the Timer's stop() method is called or the Runnable's run() method returns false.
=end

    # attr_reader :pause_length, :runnable, :running

    def initialize( pause_length, runnable )
        @pause_length = pause_length
        @runnable = runnable
        @running = false
    end

    def start()
        @id = UI.start_timer( @pause_length, true ) { 
            @running =  @runnable.run()
            stop() unless @running
        }
    end

    def stop()
        UI.stop_timer( @id )
        @running = false
    end

end # of class Timer

# test code

# class Mover

    # def initialize ( comp )
        # @comp = comp
    # end
    
    # def move( comp, r, g, b )
        # trans = comp.transformation().to_a
        # trans[12]+=r; trans[13]+=g; trans[14]+=b
        # comp.move!( Geom::Transformation.new(trans) )
        # Sketchup.active_model().active_view().invalidate()
    # end

# end # of class Mover

# class M1 < Mover
    # def run( frame )
        # move( @comp, -1, 1, 1 ) if frame < 84
        # move( @comp, 1, -1, -1 ) if frame >= 84 
    # end # of run()
    
# end # of class M1

# class M2 < Mover
    # def run( frame )
        # move( @comp, 0, 1, 1 ) if frame < 108
        # move( @comp, 0, -1, -1 ) if frame >= 108
    # end # of run()
    
# end # of class M1

# class M3 < Mover
    # def run( frame )
        # move( @comp, 1, 1, 1 ) if frame < 132
        # move( @comp,  -1, -1, -1 ) if frame >= 132
    # end # of run()
    
# end # of class M1

# model = Sketchup.active_model
# sels = model.selection
# c1 = sels[0]
# c2 = sels[1]
# c3 = sels[2]

# m1 = M1.new( c1 )
# m2 = M2.new( c2 )
# m3 = M3.new( c3 )

# Conductor.setup( 24, 8 )
# Conductor.register( m1, 1, 5 )
# Conductor.register( m2, 2, 6 )
# Conductor.register( m3, 3, 7 )
# begin
    # Conductor.start()
# rescue => e
    # puts e
# end

# end of sketchtalk_classes.rb